try:
    from .__version__ import VERSION
except:               # pragma: no cover
    VERSION='unknown'

from .file import File
from collections import OrderedDict

_EXTENSION2CLASS = OrderedDict()
def _add_extension(k,v):
    global _EXTENSION2CLASS
    d = dict(_EXTENSION2CLASS)
    d[k] = v
    _EXTENSION2CLASS = OrderedDict(sorted(d.items(),key=lambda x: len(x[0]),reverse=True))

def register_extension(ext,cls=None):
    def _register_extension(c):
        if not issubclass(c,File):
            raise TypeError("Class '%s' must extend '%s'." % (c.__name__,File.__name__))
        # _EXTENSION2CLASS[ext] = c
        _add_extension(ext,c)
        return c
    if cls is None:
        return _register_extension
    return _register_extension(cls)

from . import formats

def supports(ext=None):
    if ext is None:
        return list(_EXTENSION2CLASS.keys())
    return ext in _EXTENSION2CLASS

def AutoFile(filename,*args,**kwargs):
    fn = str(filename)
    for ext,cls in _EXTENSION2CLASS.items():
        if fn.endswith(ext):
            break
    else:
        raise KeyError("Unsupported extension for '%s'." % fn)
    return cls(filename,*args,**kwargs)

__all__ = ['File','register_extension','supports','AutoFile','formats']



__all__ = ['YamlFile']
try:
    from yaml.cyaml import CSafeLoader as Loader, CSafeDumper as Dumper
except ImportError: # pragma: no cover
    try:
        from yaml import SafeLoader as Loader, SafeDumper as Dumper
    except ImportError:
        __all__ = []

if __all__:
    import yaml
    from collections import OrderedDict
    from . import _dict_update,_flatten,File,register_extension

    # def dict_representer(dumper,data):
        # return dumper.represent_dict(data.items())
    # Dumper.add_representer(OrderedDict,dict_representer)

    @register_extension('.yml')
    @register_extension('.yaml')
    class YamlFile(File):
        def _dump(self,obj,stream=None):
            with stream or self.open('w') as cfgfile:
                yaml.dump(_flatten(obj),Dumper=Dumper,
                        stream=cfgfile,default_flow_style=False,
                        indent=2)
        def _load(self,obj=None,stream=None):
            with stream or self.open('r') as cfgfile:
                out = yaml.load(cfgfile,Loader=Loader)
            return _dict_update(obj or {}, out)



__all__=['IniFile']
try:
    import configparser
except ImportError: # pragma: no cover
    __all__=[]

if __all__:
    from . import _dict_update,_flatten,File,register_extension
    from io import TextIOWrapper,BytesIO

    def cp2dict(cp):
        return dict(filter(lambda x: len(x[1]),
                    map(lambda x: (x[0],dict(x[1])),
                        cp.items())))
    def any2cp(obj):
        if obj is None:
            return configparser.ConfigParser()
        if isinstance(obj,configparser.ConfigParser): # pragma: no cover
            return obj
        cp = configparser.ConfigParser()
        cp.read_dict(obj)
        return cp

    @register_extension('.ini')
    class IniFile(File):
        def _dump(self,obj,stream=None):
            cfg = any2cp(_flatten(obj))
            with stream or self.open('w') as cfgfile:
                cfg.write(cfgfile)
        def _load(self,obj=None,stream=None):
            cfg = any2cp(obj)
            with stream or self.open('r') as cfgfile:
                cfg.read_file(cfgfile)
            return cp2dict(cfg)



__all__=['JsonFile']
try:
    import json
except ImportError: # pragma: no cover
    __all__=[]

if __all__:
    from . import _dict_update,_flatten,File,register_extension
    from io import TextIOWrapper,BytesIO

    @register_extension('.js')
    @register_extension('.json')
    class JsonFile(File):
        def _dump(self,obj,
                stream=None,separators=(',',': '),
                indent=2,newline=True):
            with stream or self.open('w') as cfgfile:
                json.dump(_flatten(obj),cfgfile,
                        separators=separators,indent=indent)
                if newline:
                    cfgfile.write('\n')
        def _load(self,obj=None,stream=None):
            io = TextIOWrapper(BytesIO())
            with stream or self.open('r') as cfgfile:
                size = io.write(cfgfile.read())
            out = {}
            if size:
                io.seek(0)
                out = json.load(io)
            return _dict_update(obj or {}, out)



from .. import File,register_extension
import contextlib

def _flatten(obj):
    if obj is None:
        return obj
    if isinstance(obj,(str,int,float,bool)):
        return obj
    if isinstance(obj,list):
        return list(map(_flatten,obj))
    if hasattr(obj,'items'):
        return dict(map(lambda x: (x[0],_flatten(x[1])), obj.items()))
    if hasattr(obj,'keys'):
        return dict(map(lambda x: (x,_flatten(obj[x])), obj.keys()))
    return str(obj)

def _dict_update(d,s):
    if hasattr(s,'keys'):
        for k in s.keys():
            d[k] = _dict_update(d.get(k,{}),s[k])
        return d
    return s

__all__ = []

with contextlib.suppress(ImportError):
    from .js import JsonFile
    __all__.append('JsonFile')
with contextlib.suppress(ImportError):
    from .yml import YamlFile
    __all__.append('YamlFile')
with contextlib.suppress(ImportError):
    from .ini import IniFile
    __all__.append('IniFile')



# raise ValueError('Please configure setup.py first.')
PACKAGE_NAME='nzconf'
PACKAGE_DESCRIPTION='Configuration files manager'
PACKAGE_CLASSIFIERS=[
    "Programming Language :: Python :: 3",
    "Development Status :: 4 - Beta",
]
ENTRY_POINTS={
    'console_scripts': [],
}
PACKAGE_AUTHOR='Clement'
AUTHOR_EMAIL='neze+pypi@melix.org'

if __name__=='__main__':
    import setuptools,sys,os.path
    try:
        from sphinx.setup_command import BuildDoc
        cmdclass = { 'doc': BuildDoc, 'cov': BuildDoc }
    except ImportError:
        cmdclass = {}
    vstring = None
    if os.path.exists('version.py'):
        try:
            import version
            vstring = version.vstring
            rstring = version.rstring
        except ImportError:
            pass
    if vstring is None:
        try:
            from importlib import import_module
            version = import_module('{}.__version__'.format(PACKAGE_NAME))
            rstring = version.VERSION
        except ImportError:
            rstring = 'unknown'
        vstring = rstring
    doc_opt = {
        'project': ('setup.py',PACKAGE_NAME),
        'version': ('setup.py',vstring),
        'release': ('setup.py',rstring),
        'source_dir': ('setup.py','docs'),
    }
    cov_opt = dict(doc_opt)
    cov_opt.update({
        'builder': ('setup.py','coverage'),
    })
    if len(sys.argv) < 2:
        dist = 'dist'
        wheel = '-'.join([
            PACKAGE_NAME.replace('-','_'),
            rstring,
            'py3','none','any']) + '.whl'
        print(os.path.join(dist,wheel))
        sdist = '-'.join([
            PACKAGE_NAME,
            rstring]) + '.tar.gz'
        print(os.path.join(dist,sdist))
    else:
        with open('README.rst','r') as f:
            long_description = f.read()
        with open('requirements.txt','r') as f:
            install_requires = list(
                                filter(lambda x: x,
                                map(lambda x: x.strip(),
                                    f.readlines()
                                )))
        setuptools.setup(
            name=PACKAGE_NAME,
            version=rstring,
            author=PACKAGE_AUTHOR,
            author_email=AUTHOR_EMAIL,
            description=PACKAGE_DESCRIPTION,
            long_description=long_description,
            long_description_content_type='text/x-rst',
            url='',
            packages=setuptools.find_packages(),
            data_files=['requirements.txt'],
            classifiers=PACKAGE_CLASSIFIERS,
            entry_points=ENTRY_POINTS,
            install_requires=install_requires,
            cmdclass=cmdclass,
            command_options={
                'doc': doc_opt,
                'cov': cov_opt,
            }
        )



import collections
import os,os.path
from pathlib2 import Path

_mode2access = collections.OrderedDict([
    ('r',os.R_OK),('w',os.W_OK),('x',os.X_OK)])
def mode2access(mode):
    access = 0
    for m in mode:
        access |= _mode2access.get(m.lower(),0)
    return access

def access2mode(access):
    mode = ''
    for m in _mode2access.keys():
        mode += m if (access & _mode2access[m]) else '-'
    return mode

class File(object):
    def __init__(self,filename=None,mode='r'):
        self.__filename = None if filename is None else Path(filename)
        self.__access   = self.check_mode(mode)
        self.__file     = None
        self.__io       = None

    @property
    def filename(self):
        if self.__filename is not None:
            return self.__filename

    @property
    def access(self):
        return int(self.__access)
    def check_mode(self,mode):
        wanted = mode2access(mode)
        if wanted & (os.R_OK|os.W_OK) != wanted:
            raise ValueError('check_mode() only supports read/write check')
        f = self.filename
        if f is None:
            return wanted
        d = f.absolute().parent
        if not d.exists():
            raise FileNotFoundError("No such directory: '%s'" % d)
        if not d.is_dir():
            raise TypeError("Not a directory: '%s'" % d)
        if not os.access(str(d),os.X_OK):
            raise PermissionError("Cannot enter directory: '%s'" % d)
        if f.exists():
            if not f.is_file():
                raise TypeError("Not a file: '%s'" % f)
            if not os.access(str(f),wanted):
                raise PermissionError(
                        "No '%s' rights: '%s'" % (access2mode(wanted),f))
            return wanted
        if (wanted & os.W_OK) and not os.access(str(d),os.W_OK):
            raise PermissionError("Directory not writable: '%s'" % d)
        return wanted

    @property
    def opened(self):
        return self.__file is not None

    @property
    def closed(self):
        return self.__io is None or self.__io.closed

    def open(self,mode):
        if not self.opened:
            access = mode2access(mode)
            if (access & self.access) != access:
                raise PermissionError(
                        "Permission denied. Requested %s, available %s." %
                        (access2mode(access),access2mode(self.access)))
            if self.filename is None:
                raise AttributeError('No filename specified.')
            try:
                self.__file = self.filename.open(mode)
            except FileNotFoundError as e:
                self.__file = open(os.devnull,mode)
        return self
    def close(self):
        if not self.closed:
            self.__io.close()
        self.__io = None
        self.__file = None

    def __enter__(self):
        if self.closed:
            assert self.opened
            self.__io = self.__file.__enter__()
        return self
    def __exit__(self,type,value,traceback):
        self.close()

    def writable(self,*args,**kwargs):
        return (not self.closed) and self.__io.writable(*args,**kwargs)
    def write(self,*args,**kwargs):
        return self.__io.write(*args,**kwargs)

    def readable(self,*args,**kwargs):
        return (not self.closed) and self.__io.readable(*args,**kwargs)
    def read(self,*args,**kwargs):
        return self.__io.read(*args,**kwargs)

    def __iter__(self):
        return iter(self.__io)

    def dump(self,*args,**kwargs):
        return self._dump(*args,**kwargs)
    def load(self,*args,**kwargs):
        return self._load(*args,**kwargs)

    def __str__(self):
        attributes = []
        if self.opened:
            attributes.append('opened')
        if self.closed:
            attributes.append('closed')
        if self.writable():
            attributes.append('writable')
        if self.readable():
            attributes.append('readable')
        attributes = ','.join(attributes)
        if attributes:
            attributes = ',' + attributes
        return "%s('%s',mode='%s'%s)" % (type(self).__name__,str(self.filename),access2mode(self.access),attributes)
